require File.expand_path(File.dirname(__FILE__) + '/spec_helper')

describe PostcodeInfo::Postcode do
  let(:api_client){ double('api client', make_request: nil) }
  let(:postcode){ described_class.new('SN51NB', api_client) }

  describe 'constructing a Postcode' do
    let(:postcode){ 'SN51NB' }

    context 'that is valid' do
      context 'and does not contain whitespace' do
        let(:postcode){ 'YO165RB' }

        it 'does not raise an error' do
          expect{described_class.new(postcode, api_client)}.to_not raise_error
        end
      end
      context 'and contains whitespace' do
        let(:postcode){ 'N2 8AS'}

        it 'does not raise an error' do
          expect{described_class.new(postcode, api_client)}.to_not raise_error
        end
      end

      it 'does not make an API request' do
        expect( api_client ).to_not receive(:make_request)
        described_class.new(postcode, api_client)
      end

      it 'stores a normalised version of the given postcode ' do
        expect(described_class.new(" AB12  4DE\n", api_client).normalised).to eq('ab124de')
      end
    end

    context 'that is invalid' do
      let(:postcode){ 'd0nk 3y5' }

      it 'does not raise an error' do
        expect{described_class.new(postcode, api_client)}.to_not raise_error
      end
    end

  end

  describe 'valid?' do
    let(:postcode){ described_class.new('SN51NB', api_client) }

    context 'when the instance variable is nil' do
      before do
        postcode.send(:instance_variable_set, '@valid', nil)
      end

      it 'performs a lookup on the API' do
        expect( api_client ).to receive(:valid?)
        postcode.valid?
      end

      it 'stores the return value from the API in the instance variable' do
        allow(api_client).to receive(:valid?).and_return('foo')
        expect(postcode.valid?).to eq('foo')
      end
    end
        
    context 'when the instance variable has a value' do
      before do
        postcode.send(:instance_variable_set, '@valid', 'bar')
      end
      
      it 'does not perform a lookup on the API' do
        expect( api_client ).to_not receive(:valid?)
        postcode.valid?
      end

      it 'returns the existing value' do
        expect(postcode.valid?).to eq('bar')
      end
    end
  end


  describe 'addresses' do
    let(:postcode){ described_class.new('SN51NB', api_client) }

    context 'when the instance variable is nil' do
      before do
        postcode.send(:instance_variable_set, '@addresses', nil)
      end

      it 'performs a lookup on the API' do
        expect( api_client ).to receive(:addresses)
        postcode.addresses
      end

      it 'stores the return value from the API in the instance variable' do
        allow(api_client).to receive(:addresses).and_return('foo')
        expect(postcode.addresses).to eq('foo')
      end
    end
        
    context 'when the instance variable has a value' do
      before do
        postcode.send(:instance_variable_set, '@addresses', 'bar')
      end
      
      it 'does not perform a lookup on the API' do
        expect( api_client ).to_not receive(:addresses)
        postcode.addresses
      end

      it 'returns the existing value' do
        expect(postcode.addresses).to eq('bar')
      end
    end
  end

  [:longitude, :latitude].each do |method|
    describe method do
      let(:postcode){ PostcodeInfo::Postcode.new('SN51NB', api_client) }
      before do
        postcode.send(:instance_variable_set, "@#{method}", method.to_s)
      end

      context 'when the @info instance variable is nil' do
        before do
          postcode.send(:instance_variable_set, '@info', nil)
        end

        it 'performs a lookup on the API' do
          expect( postcode ).to receive(:lookup_info!)
          postcode.send(method)
        end
      end
          
      context 'when the @info instance variable has a value' do
        before do
          postcode.send(:instance_variable_set, '@info', 'info')
        end
        
        it 'does not perform a lookup on the API' do
          expect( postcode ).to_not receive(:lookup_info!)
          postcode.send(method)
        end
      end

      it "returns the value of @#{method}" do
        allow(postcode).to receive(:lookup_info!)
        expect(postcode.send(method)).to eq(method.to_s)
      end
    end
  end

  describe "lookup_info!" do
    before do
      allow(api_client).to receive(:info).and_return('info')
    end

    it "gets info from the client" do
      expect(api_client).to receive(:info)      
      allow(postcode).to receive(:parse_info!)
      postcode.lookup_info!
    end

    it "stores the info returned from the client" do
      allow(postcode).to receive(:parse_info!)
      postcode.lookup_info!
      expect(postcode.send(:instance_variable_get, '@info')).to eq('info')
    end

    it 'parses the info' do
      expect(postcode).to receive(:parse_info!)
      postcode.lookup_info!
    end
  end

  describe "parse_info!" do
    let(:stored_info){ {key: 'value', local_authority:'local authority', country: 'country'} }

    before do
      postcode.send(:instance_variable_set, '@info', stored_info )
    end

    it 'sets coordinates with @info' do
      expect(postcode).to receive(:set_coordinates!).with(stored_info)
      postcode.send(:parse_info!)
    end

    it 'stores the local_authority in @local_authority' do
      postcode.send(:parse_info!)
      expect(postcode.local_authority).to eq('local authority')
    end

    it 'stores the country in @country' do
      postcode.send(:parse_info!)
      expect(postcode.country).to eq('country')
    end

    it 'sets the valid flag' do
      expect{postcode.send(:parse_info!)}.to change{postcode.send(:instance_variable_get, '@valid')}.to(true)
    end
  end

  describe 'set_coordinates!' do
    context 'when the given info has a :centre key' do
      let(:info){ {centre: {
                      type: 'Point',
                      coordinates:[1.0, 256.8]
                    } 
                  }
                }

      it 'stores the second coordinate as @latitude' do
        postcode.send(:set_coordinates!, info)
        expect(postcode.send(:instance_variable_get, '@latitude')).to eq(1.0)
      end

      it 'stores the first coordinate as @longitude' do
        postcode.send(:set_coordinates!, info)
        expect(postcode.send(:instance_variable_get, '@longitude')).to eq(256.8)
      end
    end

    context 'when the given info does not have :centre' do
      let(:info){ {} }

      it 'does not change @latitude' do
        expect{ postcode.send(:set_coordinates!, info) }.to_not change{ postcode.send(:instance_variable_get, '@latitude') }
      end
      it 'does not change @longitude' do
        expect{ postcode.send(:set_coordinates!, info) }.to_not change{ postcode.send(:instance_variable_get, '@longitude') }
      end

    end
  end
end
